/*
replzip.c - modified from Gilles Vollant's minizip by Vesely on 29 nov 2004
Copyright (C) 2004-2007 Vesely Alessandro

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <errno.h>
#include <fcntl.h>

#ifdef unix
# include <unistd.h>
# include <utime.h>
# include <sys/types.h>
# include <sys/stat.h>
#else
# include <direct.h>
# include <io.h>
#endif

#include <ctype.h>

#include "zip.h"

#ifdef WIN32
#define USEWIN32IOAPI
#include "iowin32.h"
#endif



#define WRITEBUFFERSIZE (16384)
#if defined PATH_MAX
#define MAXFILENAME PATH_MAX
#elif defined MAX_PATH
#define MAXFILENAME MAX_PATH
#else
#define MAXFILENAME 1024
#endif

#ifdef WIN32
uLong filetime(char *f, tm_zip *tmzip, uLong* dt)
{
	int ret = 0;
	{
      FILETIME ftLocal;
      HANDLE hFind;
      WIN32_FIND_DATA  ff32;

      hFind = FindFirstFile(f,&ff32);
      if (hFind != INVALID_HANDLE_VALUE)
      {
			FileTimeToLocalFileTime(&(ff32.ftLastWriteTime),&ftLocal);
			FileTimeToDosDateTime(&ftLocal,((LPWORD)dt)+1,((LPWORD)dt)+0);
			FindClose(hFind);
			ret = 1;
      }
	}
	return ret;
}
#else
#ifdef unix
uLong filetime(char *f, tm_zip *tmzip, uLong* dt)
{
	int ret=0;
	struct stat s;        /* results of stat() */
	struct tm* filedate;
	time_t tm_t=0;

	if (strcmp(f,"-")!=0)
	{
		char name[MAXFILENAME+1];
		int len = strlen(f);

		strncpy(name, f,MAXFILENAME-1);
		/* strncpy doesnt append the trailing NULL, of the string is too long. */
		name[ MAXFILENAME ] = '\0';

		if (name[len - 1] == '/')
			name[len - 1] = '\0';
		/* not all systems allow stat'ing a file with / appended */
		if (stat(name,&s)==0)
		{
			tm_t = s.st_mtime;
			ret = 1;
		}
	}
	filedate = localtime(&tm_t);

	tmzip->tm_sec  = filedate->tm_sec;
	tmzip->tm_min  = filedate->tm_min;
	tmzip->tm_hour = filedate->tm_hour;
	tmzip->tm_mday = filedate->tm_mday;
	tmzip->tm_mon  = filedate->tm_mon ;
	tmzip->tm_year = filedate->tm_year;

	return ret;
}
#else
uLong filetime(char *f, tm_zip *tmzip, uLong* dt)
{
	return 0;
}
#endif
#endif

int check_exist_file(filename)
	const char* filename;
{
	FILE* ftestexist;
	int ret = 1;
	ftestexist = fopen(filename,"rb");
	if (ftestexist==NULL)
		ret = 0;
	else
		fclose(ftestexist);
	return ret;
}

void do_banner(void)
{
	printf("replzip 1.01, using Zlib version %s by Gilles Vollant\n",
			 ZLIB_VERSION);
}

void do_help(void)
{
	puts(
		"Usage : \n"
		"  replzip [-C \"global comment\"] file[.zip] [file_entry]...\n"
		"where\n"
		"  file_entry: [-0 to -9] [-p path] [-c comment] file...\n"
		"     -p    Set relative path for following file(s)\n"
		"     -c    Set file comment for next file\n"
		"     -0    Store only\n" \
		"     -1    Compress faster\n" \
		"     -9    Compress better\n"
		"     file  is either a file whose content is stored and base name kept,\n"
		"           or @list, where list is a file containing names of files to store:\n"
		"              * one file name per line,\n"
		"              * empty lines and lines beginning with # are discarded,\n"
		"              * lines beginning with ## are stored if no -c comment is given.\n");
}

static int opt_compress_level=Z_DEFAULT_COMPRESSION;
static char filename_try[MAXFILENAME+16];

static int add_file_in_zip(
	char *arg,          /* filespec of content     */
	char *file_path,    /* path stored in zip file */
	char *file_comment, /* comment                 */
	zipFile zf)
{
	FILE * fin = NULL;
	int size_read;
	char* filenameinzip;
	zip_fileinfo zi;
	int err = 0;
   int const size_buf=WRITEBUFFERSIZE;
   void* buf=malloc(size_buf);

   if (buf==NULL)
   {
      printf("Error allocating memory\n");
      return 3;
   }

	// there is no abs path in zip files: use "unzip -o file -d c:\" for that
	filenameinzip = strchr(file_path, ':');
	if (filenameinzip)
		file_path = filenameinzip + 1;

	while (file_path[0] == '/' || file_path[0] == '\\')
		++file_path;

	// base name
	filenameinzip = strrchr(arg, '/');
	if (filenameinzip == NULL)
	{
		filenameinzip = strrchr(arg, '\\');
		if (filenameinzip == NULL)
			filenameinzip = strchr(arg, ':');
	}
	if (filenameinzip == NULL)
		filenameinzip = arg;
	else
		++filenameinzip;

	// join path + base name
	if ((size_read = strlen(file_path)) +
		 strlen(filenameinzip) >= MAXFILENAME)
	{
		err = 1;
		printf("name too long: %s + %s\n",
				 file_path, filenameinzip);
	}
	else
	{
		char *p;
		int ch;

		if (size_read > 0)
		{
			memcpy(filename_try, file_path, size_read);
			--size_read;
			if (filename_try[size_read] != '/' &&
				 filename_try[size_read] != '\\')
				filename_try[++size_read] = '/';
			++size_read;
		}
		strcpy(filename_try + size_read, filenameinzip);
		filenameinzip = filename_try;
		for (p = filenameinzip; (ch = *p) != 0; ++p)
		{
			if (ch == '\\')
				*p = '/';
		}
	}

	memset(&zi, 0, sizeof zi);
	if (filetime(arg, &zi.tmz_date, &zi.dosDate) == 0)
	{
		printf("cannot get date of %s\n", arg);
		err = 1;
	}

	if (err == 0) err = zipOpenNewFileInZip3(
		zf,
		filenameinzip,         // full path
		&zi,                   // zip_fileinfo
		NULL,                  // extrafield_local
		0,                     // size of extrafield_local
		NULL,                  // extrafield_global
		0,                     // size of extrafield_global
		file_comment,          // comment
		(opt_compress_level != 0) ? Z_DEFLATED : 0, // method
		opt_compress_level,    // level
		0,                     // raw
		-MAX_WBITS,            // windowBits
		DEF_MEM_LEVEL,         // memlevel
		Z_DEFAULT_STRATEGY,    // strategy
		NULL,                  // password, if using encription
		0L                     // crc, if using encription
														  );

	if (err != ZIP_OK)
		printf("error in opening %s in zipfile\n",filenameinzip);
	else
	{
		fin = fopen(arg, "rb");
		if (fin==NULL)
		{
			err=ZIP_ERRNO;
			printf("error in opening %s for reading\n", arg);
		}
	}

	if (err == ZIP_OK)
	{
		printf("adding %s\n", filenameinzip);
		do
		{
			size_read = (int)fread(buf,1,size_buf,fin);
			if (size_read < size_buf && feof(fin)==0)
			{
				printf("error in reading %s\n", arg);
				err = ZIP_ERRNO;
			}

			if (size_read>0)
			{
				err = zipWriteInFileInZip (zf,buf,size_read);
				if (err<0)
				{
					printf("error in writing %s in the zipfile\n",
							 arg);
				}

			}
		} while (err == ZIP_OK && size_read>0);
	}

	if (fin)
		fclose(fin);

	if (err<0)
		err=ZIP_ERRNO;
	else
	{
		err = zipCloseFileInZip(zf);
		if (err!=ZIP_OK)
			printf("error in closing %s in the zipfile\n",
					 filenameinzip);
	}

	free(buf);
	return err;
}

int main(int argc, char *argv[])
{
   int i;
   int err=0;
   zipFile zf = NULL;
   char *global_comment = NULL;
   char *zipfname = NULL;

   if (argc==1)
   {
      do_banner();
      do_help();
      return 0;
   }
   else
   {
      char *file_comment = NULL, *file_path = "";
      zlib_filefunc_def ffunc;

#ifdef USEWIN32IOAPI
      fill_win32_filefunc(&ffunc);
#else
      fill_fopen_filefunc(&ffunc);
#endif

      for (i=1; i < argc && err == 0; ++i)
      {
         char *arg = argv[i];
         if ((*arg) == '-')
         {
            const char *p=arg+1;

            while ((*p)!='\0')
            {
               char c=*(p++);;
               if ((c>='0') && (c<='9'))
                  opt_compress_level = c-'0';
               else if (c == 'C' && i + 1 < argc)
                  global_comment = argv[++i];
               else if (c == 'c' && i + 1 < argc)
                  file_comment = argv[++i];
               else if (c == 'p' && i + 1 < argc)
                  file_path = argv[++i];
            }
         }
         else if (zf == NULL)
         {
            int i,len;
            int dot_found=0;

            strncpy(filename_try, arg, MAXFILENAME-1);
            /* strncpy doesnt append the trailing NULL, of the string is too long. */
            filename_try[ MAXFILENAME ] = '\0';

            len=(int)strlen(filename_try);
            for (i=0;i<len;i++)
               if (filename_try[i]=='.')
                  dot_found=1;

            if (dot_found==0)
               strcat(filename_try,".zip");
            zf = zipOpen2(filename_try,
								  APPEND_STATUS_CREATE, NULL, &ffunc);
            if (zf == NULL)
            {
               printf("error opening %s\n",filename_try);
               err= ZIP_ERRNO;
            }
            else if ((zipfname = strdup(filename_try)) == NULL)
				{
               err= ZIP_INTERNALERROR;
				}
				else
            {
               printf("creating %s\n", zipfname);
            }
         }
			else if (arg[0] == '@')
			{
				FILE *fplist = fopen(&arg[1], "rt");
				if (fplist)
				{
					char *list_comment = NULL;
					
					while (err == 0 && fgets(filename_try, MAXFILENAME, fplist))
					{
						char *s = &filename_try[0], *p;
						int ch;
						size_t len;

						// trim
						filename_try[MAXFILENAME] = 0;
						while ((ch = *s) != 0 && isspace(ch))
							++s;

						len = strlen(s);
						for (p = s + len; p > s;)
							if (isspace(*--p))
							{
								*p = 0;
								--len;
							}
							else
								break;
						
						// collect ## comments if no file_comment provided
						if (*s == '#')
						{
							if (file_comment == NULL &&
								 s[1] == '#' &&
								 s + 2 <= p)
							{
								s += 2;
								while (s < p && isspace(*s))
									++s;

								if (list_comment == NULL)
									list_comment = strdup(s);
								else
								{
									size_t prev_len = strlen(list_comment);
									list_comment = realloc(list_comment, prev_len + len);
									if (list_comment)
									{
										strcat(list_comment + prev_len, "\n");
										strcat(list_comment + prev_len, s);
									}
								}

								if (list_comment == NULL)
									err = 3;
							}
						}
						else
						{

							// discard empty lines
							if (len > 0)
							{
								char *fname = strdup(s);
								
								if (fname == NULL)
									err = 3;
								else
								{
									err = add_file_in_zip(
										fname, file_path,
										list_comment? list_comment: file_comment,
										zf);
									free(fname);
								}
								free(list_comment);
								list_comment = NULL;
							}
						}
					}

					free(list_comment);
					fclose(fplist);
				}
			}
         else
         {
				err = add_file_in_zip(arg, file_path, file_comment, zf);
				file_comment = NULL;
         }
      }
   }


   if (zf)
   {
      int errclose = zipClose(zf, global_comment);
      if (errclose != ZIP_OK)
      {
         err = 1;
         printf("error in closing %s\n",zipfname);
      }
   }

   if (zipfname)
   {
      if (err)
         unlink(zipfname);
      free(zipfname);
   }

   return err ? 1 : 0;
}
